/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.apache.druid.server;

import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.inject.Inject;
import org.apache.druid.client.CachingClusteredClient;
import org.apache.druid.client.cache.Cache;
import org.apache.druid.client.cache.CacheConfig;
import org.apache.druid.java.util.emitter.service.ServiceEmitter;
import org.apache.druid.query.FluentQueryRunnerBuilder;
import org.apache.druid.query.PostProcessingOperator;
import org.apache.druid.query.Query;
import org.apache.druid.query.QueryRunner;
import org.apache.druid.query.QuerySegmentWalker;
import org.apache.druid.query.QueryToolChest;
import org.apache.druid.query.QueryToolChestWarehouse;
import org.apache.druid.query.ResultLevelCachingQueryRunner;
import org.apache.druid.query.RetryQueryRunner;
import org.apache.druid.query.RetryQueryRunnerConfig;
import org.apache.druid.query.SegmentDescriptor;
import org.apache.druid.server.initialization.ServerConfig;
import org.joda.time.Interval;

/**
 */
public class ClientQuerySegmentWalker implements QuerySegmentWalker {
    private final ServiceEmitter emitter;
    private final CachingClusteredClient baseClient;
    private final QueryToolChestWarehouse warehouse;
    private final RetryQueryRunnerConfig retryConfig;
    private final ObjectMapper objectMapper;
    private final ServerConfig serverConfig;
    private final Cache cache;
    private final CacheConfig cacheConfig;


    @Inject
    public ClientQuerySegmentWalker(
            ServiceEmitter emitter,
            CachingClusteredClient baseClient,
            QueryToolChestWarehouse warehouse,
            RetryQueryRunnerConfig retryConfig,
            ObjectMapper objectMapper,
            ServerConfig serverConfig,
            Cache cache,
            CacheConfig cacheConfig
    ) {
        this.emitter = emitter;
        this.baseClient = baseClient;
        this.warehouse = warehouse;
        this.retryConfig = retryConfig;
        this.objectMapper = objectMapper;
        this.serverConfig = serverConfig;
        this.cache = cache;
        this.cacheConfig = cacheConfig;
    }

    @Override
    public <T> QueryRunner<T> getQueryRunnerForIntervals(Query<T> query, Iterable<Interval> intervals) {
        return makeRunner(query, baseClient.getQueryRunnerForIntervals(query, intervals));
    }

    //tzpdo segmentWalker是负责组织runner的, 决定了runner的逻辑
    //tzpdo 比如说broker和historical都是执行query, 组装的runner就不同
    @Override
    public <T> QueryRunner<T> getQueryRunnerForSegments(Query<T> query, Iterable<SegmentDescriptor> specs) {
        return makeRunner(
                query,
                baseClient.getQueryRunnerForSegments(query, specs)
        );
    }

    //tzpdo 这两个makeRunner又包装了几层runner
    private <T> QueryRunner<T> makeRunner(Query<T> query, QueryRunner<T> baseClientRunner) {
        QueryToolChest<T, Query<T>> toolChest = warehouse.getToolChest(query);

        // This does not adhere to the fluent workflow. See https://github.com/apache/incubator-druid/issues/5517
        return new ResultLevelCachingQueryRunner<>(makeRunner(query, baseClientRunner, toolChest),
                toolChest,
                query,
                objectMapper,
                cache,
                cacheConfig);
    }

    private <T> QueryRunner<T> makeRunner(
            Query<T> query,
            QueryRunner<T> baseClientRunner,
            QueryToolChest<T, Query<T>> toolChest
    ) {
        PostProcessingOperator<T> postProcessing = objectMapper.convertValue(
                query.<String>getContextValue("postProcessing"),
                new TypeReference<PostProcessingOperator<T>>() {
                }
        );

        return new FluentQueryRunnerBuilder<>(toolChest)
                .create(
                        new SetAndVerifyContextQueryRunner<>(
                                serverConfig,
                                //tzpdo 由于是调http请求查, 存在失败的可能性, 这里包装了一个重试的runner
                                new RetryQueryRunner<>(
                                        baseClientRunner,
                                        retryConfig,
                                        objectMapper
                                )
                        )
                )
                .applyPreMergeDecoration()
                //tzpdo 合并结果, 做aggregator和post aggregator
                .mergeResults()
                .applyPostMergeDecoration()
                .emitCPUTimeMetric(emitter)
                .postProcess(postProcessing);
    }
}
